@@ -1,7 +1,7 @@
 /*
  * libeio implementation
  *
- * Copyright (c) 2007,2008,2009,2010,2011,2012,2013,2016 Marc Alexander Lehmann <libeio@schmorp.de>
+ * Copyright (c) 2007,2008,2009,2010,2011,2012,2013,2016,2017,2018,2019 Marc Alexander Lehmann <libeio@schmorp.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without modifica-
@@ -203,11 +203,15 @@ static void eio_destroy (eio_req *req);
   symlink (const char *old, const char *neu)
   {
     #if WINVER >= 0x0600
-      if (CreateSymbolicLink (neu, old, 1))
-        return 0;
+      int flags;
 
-      if (CreateSymbolicLink (neu, old, 0))
-        return 0;
+      /* This tries out all combinations of SYMBOLIC_LINK_FLAG_DIRECTORY
+       * and SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE,
+       * with directory first.
+       */
+      for (flags = 3; flags >= 0; --flags)
+        if (CreateSymbolicLink (neu, old, flags))
+          return 0;
     #endif
 
     return EIO_ERRNO (ENOENT, -1);
@@ -229,6 +233,7 @@ static void eio_destroy (eio_req *req);
 
 #else
 
+  #include <sys/ioctl.h>
   #include <sys/time.h>
   #include <sys/select.h>
   #include <unistd.h>
@@ -297,6 +302,11 @@ static void eio_destroy (eio_req *req);
 # endif
 #endif
 
+#if HAVE_RENAMEAT2
+# include <sys/syscall.h>
+# include <linux/fs.h>
+#endif
+
 #ifndef D_TYPE
 # define D_TYPE(de) 0
 #endif
@@ -314,9 +324,31 @@ static void eio_destroy (eio_req *req);
 
 /* used for readlink etc. */
 #ifndef PATH_MAX
-# define PATH_MAX 4096
+# define PATH_MAX 0
 #endif
 
+#ifndef O_CLOEXEC
+  #define O_CLOEXEC 0
+#endif
+
+#ifndef O_NONBLOCK
+  #define O_NONBLOCK 0
+#endif
+
+#ifndef O_SEARCH
+  #define O_SEARCH O_RDONLY
+#endif
+
+#ifndef O_DIRECTORY
+  #define O_DIRECTORY 0
+#endif
+
+#ifndef EIO_PATH_MIN
+# define EIO_PATH_MIN 8160
+#endif
+
+#define EIO_PATH_MAX (PATH_MAX <= EIO_PATH_MIN ? EIO_PATH_MIN : PATH_MAX)
+
 /* buffer size for various temporary buffers */
 #define EIO_BUFSIZE 65536
 
@@ -336,9 +368,6 @@ struct etp_tmpbuf;
 #if _POSIX_VERSION >= 200809L
   #define HAVE_AT 1
   #define WD2FD(wd) ((wd) ? (wd)->fd : AT_FDCWD)
-  #ifndef O_SEARCH
-    #define O_SEARCH O_RDONLY
-  #endif
 #else
   #define HAVE_AT 0
   static const char *wd_expand (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path);
@@ -539,6 +568,21 @@ int eio_poll (void)
 /*****************************************************************************/
 /* work around various missing functions */
 
+#if HAVE_POSIX_CLOSE && !__linux
+# define eio__close(fd) posix_close (fd, 0)
+#else
+# define eio__close(fd) close (fd)
+#endif
+
+/* close() without disturbing errno */
+static void
+silent_close (int fd)
+{
+  int saved_errno = errno;
+  eio__close (fd);
+  errno = saved_errno;
+}
+
 #ifndef HAVE_UTIMES
 
 # undef utimes
@@ -860,12 +904,21 @@ eio__mlockall (int flags)
     mallopt (-6, 238); /* http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=473812 */
   #endif
 
-  if (EIO_MCL_CURRENT   != MCL_CURRENT
-      || EIO_MCL_FUTURE != MCL_FUTURE)
+  #ifndef MCL_ONFAULT
+    if (flags & EIO_MCL_ONFAULT)
+      return EIO_ERRNO (EINVAL, -1);
+    #define MCL_ONFAULT 4
+  #endif
+
+  if (EIO_MCL_CURRENT     != MCL_CURRENT
+      || EIO_MCL_FUTURE   != MCL_FUTURE
+      || EIO_MCL_ONFAULT  != MCL_ONFAULT)
     {
       flags = 0
          | (flags & EIO_MCL_CURRENT ? MCL_CURRENT : 0)
-         | (flags & EIO_MCL_FUTURE  ? MCL_FUTURE : 0);
+         | (flags & EIO_MCL_FUTURE  ? MCL_FUTURE  : 0)
+         | (flags & EIO_MCL_ONFAULT ? MCL_ONFAULT : 0)
+      ;
     }
 
   return mlockall (flags);
@@ -926,9 +979,9 @@ eio__mtouch (eio_req *req)
 
     if (addr < end)
       if (flags & EIO_MT_MODIFY) /* modify */
-        do { *((volatile sig_atomic_t *)addr) |= 0; } while ((addr += page) < len && !EIO_CANCELLED (req));
+        do { *((volatile sig_atomic_t *)addr) |= 0; } while ((addr += page) < end && !EIO_CANCELLED (req));
       else
-        do { *((volatile sig_atomic_t *)addr)     ; } while ((addr += page) < len && !EIO_CANCELLED (req));
+        do { *((volatile sig_atomic_t *)addr)     ; } while ((addr += page) < end && !EIO_CANCELLED (req));
   }
 
   return 0;
@@ -972,15 +1025,15 @@ eio__realpath (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
   if (!*rel)
     return -1;
 
-  res = etp_tmpbuf_get (tmpbuf, PATH_MAX * 3);
+  res = etp_tmpbuf_get (tmpbuf, EIO_PATH_MAX * 3);
 #ifdef _WIN32
   if (_access (rel, 4) != 0)
     return -1;
 
-  symlinks = GetFullPathName (rel, PATH_MAX * 3, res, 0);
+  symlinks = GetFullPathName (rel, EIO_PATH_MAX * 3, res, 0);
 
   errno = ENAMETOOLONG;
-  if (symlinks >= PATH_MAX * 3)
+  if (symlinks >= EIO_PATH_MAX * 3)
     return -1;
 
   errno = EIO;
@@ -990,8 +1043,8 @@ eio__realpath (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
   return symlinks;
 
 #else
-  tmp1 = res  + PATH_MAX;
-  tmp2 = tmp1 + PATH_MAX;
+  tmp1 = res  + EIO_PATH_MAX;
+  tmp2 = tmp1 + EIO_PATH_MAX;
 
 #if 0 /* disabled, the musl way to do things is just too racy */
 #if __linux && defined(O_NONBLOCK) && defined(O_NOATIME)
@@ -1002,9 +1055,9 @@ eio__realpath (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
     if (fd >= 0)
       {
         sprintf (tmp1, "/proc/self/fd/%d", fd);
-        req->result = readlink (tmp1, res, PATH_MAX);
+        req->result = readlink (tmp1, res, EIO_PATH_MAX);
         /* here we should probably stat the open file and the disk file, to make sure they still match */
-        close (fd);
+        eio__close (fd);
 
         if (req->result > 0)
           goto done;
@@ -1025,7 +1078,7 @@ eio__realpath (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
       
       if (wd == EIO_CWD)
         {
-          if (!getcwd (res, PATH_MAX))
+          if (!getcwd (res, EIO_PATH_MAX))
             return -1;
 
           len = strlen (res);
@@ -1082,7 +1135,7 @@ eio__realpath (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
         res [len + 1] = 0;
 
         /* now check if it's a symlink */
-        linklen = readlink (tmpbuf->ptr, tmp1, PATH_MAX);
+        linklen = readlink (tmpbuf->ptr, tmp1, EIO_PATH_MAX);
 
         if (linklen < 0)
           {
@@ -1098,7 +1151,7 @@ eio__realpath (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
             int rellen = strlen (rel);
 
             errno = ENAMETOOLONG;
-            if (linklen + 1 + rellen >= PATH_MAX)
+            if (linklen + 1 + rellen >= EIO_PATH_MAX) /* also catch linklen >= EIO_PATH_MAX */
               return -1;
 
             errno = ELOOP;
@@ -1362,26 +1415,26 @@ eio__scandir (eio_req *req, etp_worker *self)
   }
 #else
   #if HAVE_AT
-    if (req->wd)
-      {
-        int fd = openat (WD2FD (req->wd), req->ptr1, O_CLOEXEC | O_SEARCH | O_DIRECTORY);
+    {
+      int fd = openat (WD2FD (req->wd), req->ptr1, O_CLOEXEC | O_SEARCH | O_DIRECTORY | O_NONBLOCK);
 
-        if (fd < 0)
+      if (fd < 0)
+        return;
+
+      dirp = fdopendir (fd);
+
+      if (!dirp)
+        {
+          silent_close (fd);
           return;
-
-        dirp = fdopendir (fd);
-
-        if (!dirp)
-          close (fd);
-      }
-    else
-      dirp = opendir (req->ptr1);
+        }
+    }
   #else
     dirp = opendir (wd_expand (&self->tmpbuf, req->wd, req->ptr1));
-  #endif
 
-  if (!dirp)
-    return;
+    if (!dirp)
+      return;
+  #endif
 #endif
 
   if (req->flags & EIO_FLAG_PTR1_FREE)
@@ -1626,7 +1679,21 @@ eio__wd_open_sync (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
     return EIO_INVALID_WD;
 
 #if HAVE_AT
-  fd = openat (WD2FD (wd), path, O_CLOEXEC | O_SEARCH | O_DIRECTORY);
+  fd = openat (WD2FD (wd), path, O_CLOEXEC | O_SEARCH | O_DIRECTORY | O_NONBLOCK);
+
+  /* 0 is a valid fd, but we use it for EIO_CWD, so in the very unlikely */
+  /* case of fd 0 being available (almost certainly an a pplication bug) */
+  /* make sure we use another fd value */
+  #if EIO_CWD
+    error EIO_CWD must be 0
+  #endif
+  if (ecb_expect_false (fd == 0))
+    {
+      int fd2 = fcntl (fd, F_DUPFD_CLOEXEC ? F_DUPFD_CLOEXEC : F_DUPFD);
+      fcntl (fd2, F_SETFD, FD_CLOEXEC);
+      eio__close (fd);
+      fd = fd2;
+    }
 
   if (fd < 0)
     return EIO_INVALID_WD;
@@ -1648,7 +1715,7 @@ eio__wd_open_sync (struct etp_tmpbuf *tmpbuf, eio_wd wd, const char *path)
 eio_wd
 eio_wd_open_sync (eio_wd wd, const char *path)
 {
-  struct etp_tmpbuf tmpbuf = { };
+  struct etp_tmpbuf tmpbuf = { 0 };
   wd = eio__wd_open_sync (&tmpbuf, wd, path);
   free (tmpbuf.ptr);
 
@@ -1661,7 +1728,7 @@ eio_wd_close_sync (eio_wd wd)
   if (wd != EIO_INVALID_WD && wd != EIO_CWD)
     {
       #if HAVE_AT
-      close (wd->fd);
+      eio__close (wd->fd);
       #endif
       free (wd);
     }
@@ -1669,35 +1736,47 @@ eio_wd_close_sync (eio_wd wd)
 
 #if HAVE_AT
 
+static int
+eio__renameat2 (int olddirfd, const char *oldpath, int newdirfd, const char *newpath, unsigned int flags)
+{
+#if HAVE_RENAMEAT2
+  return syscall (SYS_renameat2, olddirfd, oldpath, newdirfd, newpath, flags);
+#else
+  if (flags)
+    return EIO_ENOSYS ();
+
+  return renameat (olddirfd, oldpath, newdirfd, newpath);
+#endif
+}
+
 /* they forgot these */
 
 static int
 eio__truncateat (int dirfd, const char *path, off_t length)
 {
-  int fd = openat (dirfd, path, O_WRONLY | O_CLOEXEC);
+  int fd = openat (dirfd, path, O_WRONLY | O_CLOEXEC | O_NONBLOCK);
   int res;
 
   if (fd < 0)
     return fd;
 
   res = ftruncate (fd, length);
-  close (fd);
+  silent_close (fd);
   return res;
 }
 
 static int
 eio__statvfsat (int dirfd, const char *path, struct statvfs *buf)
 {
-  int fd = openat (dirfd, path, O_SEARCH | O_CLOEXEC);
+  int fd = openat (dirfd, path, O_SEARCH | O_CLOEXEC | O_NONBLOCK);
   int res;
 
   if (fd < 0)
     return fd;
 
   res = fstatvfs (fd, buf);
-  close (fd);
+  silent_close (fd);
   return res;
-
 }
 
 #endif
@@ -1715,12 +1794,40 @@ eio__statvfsat (int dirfd, const char *path, struct statvfs *buf)
         {					\
           errno       = ENOMEM;			\
           req->result = -1;			\
-          break;				\
+          goto alloc_fail;			\
         }					\
     }
 
 /*****************************************************************************/
 
+static void
+eio__slurp (int fd, eio_req *req)
+{
+  req->result = fd;
+
+  if (fd < 0)
+    return;
+
+  if (req->offs < 0 || !req->size) /* do we need the size? */
+    {
+      off_t size = lseek (fd, 0, SEEK_END);
+
+      if (req->offs < 0)
+        req->offs += size;
+
+      if (!req->size)
+        req->size = size - req->offs;
+    }
+
+  ALLOC (req->size);
+  req->result = pread (fd, req->ptr2, req->size, req->offs);
+
+  silent_close (fd);
+
+alloc_fail:
+  ;
+}
+
 int ecb_cold
 eio_init (void (*want_poll)(void), void (*done_poll)(void))
 {
@@ -1827,6 +1934,7 @@ eio_execute (etp_worker *self, eio_req *req)
       case EIO_CHMOD:     req->result = fchmodat  (dirfd, req->ptr1, (mode_t)req->int2, 0); break;
       case EIO_TRUNCATE:  req->result = eio__truncateat (dirfd, req->ptr1, req->offs); break;
       case EIO_OPEN:      req->result = openat    (dirfd, req->ptr1, req->int1, (mode_t)req->int2); break;
+      case EIO_SLURP:     eio__slurp (  openat    (dirfd, req->ptr1, O_RDONLY | O_CLOEXEC), req); break;
 
       case EIO_UNLINK:    req->result = unlinkat  (dirfd, req->ptr1, 0); break;
       case EIO_RMDIR:     /* complications arise because "." cannot be removed, so we might have to expand */
@@ -1834,17 +1942,28 @@ eio_execute (etp_worker *self, eio_req *req)
                              ? rmdir (req->wd->str)
                              : unlinkat  (dirfd, req->ptr1, AT_REMOVEDIR); break;
       case EIO_MKDIR:     req->result = mkdirat   (dirfd, req->ptr1, (mode_t)req->int2); break;
-      case EIO_RENAME:    /* complications arise because "." cannot be renamed, so we might have to expand */
-                          req->result = req->wd && SINGLEDOT (req->ptr1)
-                             ? rename (req->wd->str, req->ptr2)
-                             : renameat (dirfd, req->ptr1, WD2FD ((eio_wd)req->int3), req->ptr2); break;
+      case EIO_RENAME:    req->result = eio__renameat2 (
+                            dirfd,
+                            /* complications arise because "." cannot be renamed, so we might have to expand */
+                            req->wd && SINGLEDOT (req->ptr1) ? req->wd->str : req->ptr1,
+                            WD2FD ((eio_wd)req->int3),
+                            req->ptr2,
+                            req->int2
+                          );
+                          break;
       case EIO_LINK:      req->result = linkat    (dirfd, req->ptr1, WD2FD ((eio_wd)req->int3), req->ptr2, 0); break;
       case EIO_SYMLINK:   req->result = symlinkat (req->ptr1, dirfd, req->ptr2); break;
       case EIO_MKNOD:     req->result = mknodat   (dirfd, req->ptr1, (mode_t)req->int2, (dev_t)req->offs); break;
-      case EIO_READLINK:  ALLOC (PATH_MAX);
-                          req->result = readlinkat (dirfd, req->ptr1, req->ptr2, PATH_MAX); break;
       case EIO_STATVFS:   ALLOC (sizeof (EIO_STRUCT_STATVFS));
                           req->result = eio__statvfsat (dirfd, req->ptr1, (EIO_STRUCT_STATVFS *)req->ptr2); break;
+      case EIO_READLINK:  ALLOC (EIO_PATH_MAX);
+                          req->result = readlinkat (dirfd, req->ptr1, req->ptr2, EIO_PATH_MAX);
+                          if (req->result == EIO_PATH_MAX)
+                            {
+                              req->result = -1;
+                              errno = ENAMETOOLONG;
+                            }
+                          break;
       case EIO_UTIME:
       case EIO_FUTIME:
         {
@@ -1879,18 +1998,25 @@ eio_execute (etp_worker *self, eio_req *req)
       case EIO_CHMOD:     req->result = chmod     (path     , (mode_t)req->int2); break;
       case EIO_TRUNCATE:  req->result = truncate  (path     , req->offs); break;
       case EIO_OPEN:      req->result = open      (path     , req->int1, (mode_t)req->int2); break;
+      case EIO_SLURP:     eio__slurp (  open      (path     , O_RDONLY | O_CLOEXEC), req); break;
 
       case EIO_UNLINK:    req->result = unlink    (path     ); break;
       case EIO_RMDIR:     req->result = rmdir     (path     ); break;
       case EIO_MKDIR:     req->result = mkdir     (path     , (mode_t)req->int2); break;
-      case EIO_RENAME:    req->result = rename    (path     , req->ptr2); break;
+      case EIO_RENAME:    req->result = req->int2 ? EIO_ENOSYS () : rename (path, req->ptr2); break;
       case EIO_LINK:      req->result = link      (path     , req->ptr2); break;
       case EIO_SYMLINK:   req->result = symlink   (path     , req->ptr2); break;
       case EIO_MKNOD:     req->result = mknod     (path     , (mode_t)req->int2, (dev_t)req->offs); break;
-      case EIO_READLINK:  ALLOC (PATH_MAX);
-                          req->result = readlink  (path, req->ptr2, PATH_MAX); break;
       case EIO_STATVFS:   ALLOC (sizeof (EIO_STRUCT_STATVFS));
                           req->result = statvfs   (path     , (EIO_STRUCT_STATVFS *)req->ptr2); break;
+      case EIO_READLINK:  ALLOC (EIO_PATH_MAX);
+                          req->result = readlink (path, req->ptr2, EIO_PATH_MAX);
+                          if (req->result == EIO_PATH_MAX)
+                            {
+                              req->result = -1;
+                              errno = ENAMETOOLONG;
+                            }
+                          break;
 
       case EIO_UTIME:
       case EIO_FUTIME:
@@ -1935,7 +2061,7 @@ eio_execute (etp_worker *self, eio_req *req)
       case EIO_FCHMOD:    req->result = fchmod    (req->int1, (mode_t)req->int2); break;
       case EIO_FTRUNCATE: req->result = ftruncate (req->int1, req->offs); break;
 
-      case EIO_CLOSE:     req->result = close     (req->int1); break;
+      case EIO_CLOSE:     req->result = eio__close (req->int1); break;
       case EIO_DUP2:      req->result = dup2      (req->int1, req->int2); break;
       case EIO_SYNC:      req->result = 0; sync (); break;
       case EIO_FSYNC:     req->result = fsync     (req->int1); break;
@@ -1983,6 +2109,7 @@ eio_execute (etp_worker *self, eio_req *req)
         break;
     }
 
+alloc_fail:
   req->errorno = errno;
 }
 
@@ -2245,6 +2372,11 @@ eio_req *eio_rename (const char *path, const char *new_path, int pri, eio_cb cb,
   return eio__2path (EIO_RENAME, path, new_path, pri, cb, data);
 }
 
+eio_req *eio_slurp (const char *path, void *buf, size_t length, off_t offset, int pri, eio_cb cb, void *data)
+{
+  REQ (EIO_SLURP); PATH; req->offs = offset; req->size = length; req->ptr2 = buf; SEND;
+}
+
 eio_req *eio_custom (void (*execute)(eio_req *), int pri, eio_cb cb, void *data)
 {
   REQ (EIO_CUSTOM); req->feed = execute; SEND;
@@ -2311,3 +2443,8 @@ eio_sendfile_sync (int ofd, int ifd, off_t offset, size_t count)
   return eio__sendfile (ofd, ifd, offset, count);
 }
 
+int eio_mlockall_sync (int flags)
+{
+  return eio__mlockall (flags);
+}
+
